/*
 ***********************************************************************
 * Copyright 2005-2006 by Freescale Semiconductor, Inc.
 * All modifications are confidential and proprietary information
 * of Freescale Semiconductor, Inc. ALL RIGHTS RESERVED.
 ***********************************************************************
 */

/*******************************************************************************
 *
 * Motorola Inc.
 * (c) Copyright 2004 Motorola, Inc.
 * ALL RIGHTS RESERVED
 *
 * *****************************************************************************
 * File name: aac_aux.c
 *
 * Description: Contains functions which the application needs to parse headers.
 *
 * Functions Included: App_bs_look_bits
 *                     App_FindFileType
 *                     App_bs_refill_buffer
 *                     App_bs_readinit
 *                     App_bs_read_bits
 *                     App_bs_byte_align
 *                     App_get_ele_list
 *                     App_get_prog_config
 *                     App_get_adif_header
 *                     App_get_adts_header
 *******************************************************************************
 *
 *   CHANGE HISTORY
 *   dd/mm/yy   Code Ver     Description                        Author
 *   --------   -------      -----------                        ------
 *  19/08/2004     01     Created.						            S.Nachiappan
 *                           Has functions, which the
 *                           application needs to
 *                           parse headers.
 *  29/08/2005     02       Removed aacd_tables.h header	        Ashok Kumar
 *                           file
 *  13/10/2006     03       Raw Bitstream Support Code				ShyamKrishnan M 
 *
 *	08/02/2007     11		Bitrate support for ADTS streams		Madhav Varma G P
 *
 ******************************************************************************/
#include <gst/gst.h>
#include <string.h>
#include "aacd_dec_interface.h"
#include "mfw_gst_aacdec.h"
#include "aac_aux.h"

#include <string.h>

#define BS_BUF_SIZE AACD_INPUT_BUFFER_SIZE
#define MAX_ENC_BUF_SIZE 400*BS_BUF_SIZE
#define LONG_BOUNDARY 4
#define FRAMESIZE 1024

int sampling_frequency[] = { 96000, 88200, 64000, 48000, 44100, 32000, 24000, 22050, 16000, 12000, 11025, 8000, 0, 0, 0, 0};
enum { BITSPERFRAME_SCALE = 4 } ;
unsigned int lastGoodBufferFullness,lastGoodBufferFullness,firstBufferFullness;
 
BitstreamParam App_bs_param;
 //int bitstream_count;
 char bitstream_buf[MAX_ENC_BUF_SIZE];
// int bitstream_buf_index;
// int in_buf_done;
 int bytes_supplied;
 unsigned char App_ibs_buf[INTERNAL_BS_BUFSIZE];
/***************************************************************
 *  FUNCTION NAME - App_bs_look_bits
 *
 *  DESCRIPTION
 *      Get the required number of bits from the bit-register
 *
 *  ARGUMENTS
 *      nbits - Number of bits required
 *
 *  RETURN VALUE
 *      Required bits
 *
 **************************************************************/
unsigned long App_bs_look_bits (int nbits,MFW_GST_AACDEC_INFO_T *aacdec_info)
{
   BitstreamParam *b = &(aacdec_info->app_params.bs_param);
    return b->bit_register;
}


/***************************************************************
 *  FUNCTION NAME - App_FindFileType
 *
 *  DESCRIPTION
 *      Find if the file is of type, ADIF or ADTS
 *
 *  ARGUMENTS
 *      val     -   First 4 bytes in the stream
 *
 *  RETURN VALUE
 *      0 - Success
 *     -1 - Error
 *
 **************************************************************/
int App_FindFileType(int val,MFW_GST_AACDEC_INFO_T *aacdec_info)
{

    
  if (val == 0x41444946)
    {
      aacdec_info->app_params.App_adif_header_present = TRUE;
    }
  else
    {
      if ((val & 0xFFF00000) == 0xFFF00000)
	    aacdec_info->app_params.App_adts_header_present = TRUE;
      else
	return(-1);
    }
  return(0);
}

int ADTSBitrate(AACD_Decoder_info *dec_info,
                MFW_GST_AACDEC_INFO_T *aacdec_info)
{
    AACD_Block_Params * params;
    params = &aacdec_info->app_params.BlockParams;
    unsigned int unSampleRate = sampling_frequency[aacdec_info->SampFreqIdx] ;
    GST_DEBUG("Sampling Frequency=%d\n",unSampleRate);
    const unsigned int unFrameSamples = FRAMESIZE ;
    dec_info->aacd_bit_rate =
        ((gfloat)aacdec_info->bitsPerFrame * unSampleRate)/unFrameSamples;
    GST_DEBUG("Bitrate=%d\n",dec_info->aacd_bit_rate);
    return 1 ;
}





/***************************************************************
 *  FUNCTION NAME - App_bs_refill_buffer
 *
 *  DESCRIPTION
 *      Fill the bitstream buffer with new buffer.
 *
 *  ARGUMENTS
 *      None
 *
 *  RETURN VALUE
 *      0  - success
 *      -1 - error
 *
 **************************************************************/
int App_bs_refill_buffer (MFW_GST_AACDEC_INFO_T *aacdec_info)
{
    BitstreamParam *b = &(App_bs_param);
    int bytes_to_copy;
    unsigned int len;


    bytes_to_copy = b->bs_end_ext - b->bs_curr_ext;
    if (bytes_to_copy <= 0)
    {
      if (aacdec_info->bitstream_count <=0)
	return(-1);
      else
	{
	  len = (aacdec_info->bitstream_count > BS_BUF_SIZE) ? BS_BUF_SIZE : aacdec_info->bitstream_count;

	  b->bs_curr_ext = (unsigned char *)(bitstream_buf + aacdec_info->bitstream_buf_index);
	  b->bs_end_ext = b->bs_curr_ext + len;
	  bytes_to_copy = len;

	  aacdec_info->bitstream_buf_index += len;
	  aacdec_info->bitstream_count -= len;
	  aacdec_info->in_buf_done    += len;
	  bytes_supplied = len;

	  /* Set only if previous Seeking is done */
	  /*
	    if (b->bs_seek_needed == 0)
            b->bs_seek_needed = SeekFlag;
	 */ 
	}

    }

    if (bytes_to_copy > INTERNAL_BS_BUFSIZE)
      bytes_to_copy = INTERNAL_BS_BUFSIZE;

    b->bs_curr = App_ibs_buf;
    memcpy(b->bs_curr, b->bs_curr_ext, bytes_to_copy);
    b->bs_curr_ext += bytes_to_copy;
    b->bs_end = b->bs_curr + bytes_to_copy;

    return 0;

}


/***********************************************************************
 *
 *   FUNCTION NAME - App_bs_readinit
 *
 *   DESCRIPTION
 *      This module initializes the BitStream Parameteres.
 *
 *   ARGUMENTS
 *       buf       -  Buffer from which, data is to be copied
 *                    into internal-buffer and bit-register
 *       bytes     -  Size of the above buffer in bytes.
 *
 *   RETURN VALUE
 *      None
 **********************************************************************/
void App_bs_readinit(char *buf, int bytes,MFW_GST_AACDEC_INFO_T *aacdec_info)
{

    BitstreamParam *b = &(aacdec_info->app_params.bs_param);
    unsigned int temp;
    int ret;

    b->bs_curr = (unsigned char *)buf;
    b->bs_end = b->bs_curr + bytes;
    b->bs_eof = 0;
    b->bs_seek_needed = 0;
    b->bit_register = 0;
    b->bit_counter = BIT_COUNTER_INIT;

    while (b->bit_counter >= 0)
    {
       if (b->bs_curr >= b->bs_end)
        {
            ret = App_bs_refill_buffer(aacdec_info);
            if (ret < 0)
                break;
        }

		
        temp = *b->bs_curr++;
        b->bit_register = b->bit_register | (temp << b->bit_counter);
        b->bit_counter -= 8;
    }

}

/***************************************************************
 *  FUNCTION NAME - App_bs_read_bits
 *
 *  DESCRIPTION
 *      Reads the given number of bits from the bit-register
 *
 *  ARGUMENTS
 *      nbits - Number of bits required
 *
 *  RETURN VALUE
 *      - Required bits
 *      - -1 in case of end of bitstream
 *
 **************************************************************/
unsigned long App_bs_read_bits (int nbits,MFW_GST_AACDEC_INFO_T *aacdec_info)
{
    BitstreamParam *b = &(aacdec_info->app_params.bs_param);
    unsigned long temp,temp1,temp_bit_register;
    long temp_bit_counter;
    int ret;

    aacdec_info->app_params.BitsInHeader += nbits;


    temp_bit_counter = b->bit_counter;
    temp_bit_register = b->bit_register;

    /* If more than available bits are requested,
     * return error
     */
    if ((MIN_REQD_BITS - temp_bit_counter) < nbits)
        return 0;


    temp = temp_bit_register >> (32 - nbits);
    temp_bit_register <<= nbits;
    temp_bit_counter += nbits;

    while (temp_bit_counter >= 0)
    {
        if (b->bs_curr >= b->bs_end)
        {
            ret = App_bs_refill_buffer(aacdec_info);
            if (ret < 0)
            {
                b->bit_register = temp_bit_register;
                b->bit_counter = temp_bit_counter;

                return(temp);
            }
        }

        temp1 = *b->bs_curr++;
        temp_bit_register = temp_bit_register | (temp1 << temp_bit_counter);
        temp_bit_counter -= 8;
     }

     b->bit_register = temp_bit_register;
     b->bit_counter = temp_bit_counter;


    return(temp);
}

/*******************************************************************************
 *
 *   FUNCTION NAME - App_bs_byte_align
 *
 *   DESCRIPTION
 *       This function makes the number of bits in the bit register
 *       to be a multiple of 8.
 *
 *   ARGUMENTS
 *       None
 *
 *   RETURN VALUE
 *       number of bits discarded.
*******************************************************************************/
int App_bs_byte_align(MFW_GST_AACDEC_INFO_T *aacdec_info)
{
    BitstreamParam *b = &(aacdec_info->app_params.bs_param);
    int             nbits;

    nbits = MIN_REQD_BITS - b->bit_counter;
    nbits = nbits & 0x7;    /* LSB 3 bits */

    aacdec_info->app_params.BitsInHeader += nbits;


    b->bit_register <<= nbits;
    b->bit_counter += nbits;


    return (nbits);
}


/*******************************************************************************
 *
 *   FUNCTION NAME - App_get_ele_list
 *
 *   DESCRIPTION
 *       Gets a list of elements present in the current data block
 *
 *   ARGUMENTS
 *       p           -  Pointer to array of Elements.
 *       enable_cpe  _  Flag to indicate whether channel paired element is
 *                      present.
 *
 *   RETURN VALUE
 *       None
*******************************************************************************/
static void App_get_ele_list(AACD_EleList * p, int enable_cpe,MFW_GST_AACDEC_INFO_T *aacdec_info)
{
    int             i,
                    j;

    for (i = 0, j = p->num_ele; i < j; i++)
    {
        if (enable_cpe)
            p->ele_is_cpe[i] = App_bs_read_bits(LEN_ELE_IS_CPE,aacdec_info);
        else
            p->ele_is_cpe[i] = 0;
        p->ele_tag[i] = App_bs_read_bits(LEN_TAG,aacdec_info);
    }

}



/*******************************************************************************
 *
 *   FUNCTION NAME - App_get_prog_config
 *
 *   DESCRIPTION
 *       Read the program configuration element from the data block
 *
 *   ARGUMENTS
 *       p           -  Pointer to a structure to store the new program
 *                       configuration.
 *
 *   RETURN VALUE
 *         Success :  0
 *
 *         Error   : -1
 *
*******************************************************************************/
int App_get_prog_config(AACD_ProgConfig * p,MFW_GST_AACDEC_INFO_T *aacdec_info)
{
    int             i,
                    j;


    p->tag = App_bs_read_bits(LEN_TAG,aacdec_info);

    p->profile = App_bs_read_bits(LEN_PROFILE,aacdec_info);
    if (p->profile != 1)
    {
        return -1;
    }
    p->sampling_rate_idx = App_bs_read_bits(LEN_SAMP_IDX,aacdec_info);
    if (p->sampling_rate_idx >= 0xc)
    {
        return -1;
    }
    p->front.num_ele = App_bs_read_bits(LEN_NUM_ELE,aacdec_info);
    if (p->front.num_ele > FChans)
    {
        return -1;
    }
    p->side.num_ele = App_bs_read_bits(LEN_NUM_ELE,aacdec_info);
    if (p->side.num_ele > SChans)
    {
        return -1;
    }
    p->back.num_ele = App_bs_read_bits(LEN_NUM_ELE,aacdec_info);
    if (p->back.num_ele > BChans)
    {
        return -1;
    }
    p->lfe.num_ele = App_bs_read_bits(LEN_NUM_LFE,aacdec_info);
    if (p->lfe.num_ele > LChans)
    {
        return -1;
    }
    p->data.num_ele = App_bs_read_bits(LEN_NUM_DAT,aacdec_info);
    p->coupling.num_ele = App_bs_read_bits(LEN_NUM_CCE,aacdec_info);
    if (p->coupling.num_ele > CChans)
    {
        return -1;
    }
    if ((p->mono_mix.present = App_bs_read_bits(LEN_MIX_PRES,aacdec_info)) == 1)
        p->mono_mix.ele_tag = App_bs_read_bits(LEN_TAG,aacdec_info);
    if ((p->stereo_mix.present = App_bs_read_bits(LEN_MIX_PRES,aacdec_info)) == 1)
        p->stereo_mix.ele_tag = App_bs_read_bits(LEN_TAG,aacdec_info);
    if ((p->matrix_mix.present = App_bs_read_bits(LEN_MIX_PRES,aacdec_info)) == 1)
    {
        p->matrix_mix.ele_tag = App_bs_read_bits(LEN_MMIX_IDX,aacdec_info);
        p->matrix_mix.pseudo_enab = App_bs_read_bits(LEN_PSUR_ENAB,aacdec_info);
    }
    App_get_ele_list(&p->front, 1,aacdec_info);
    App_get_ele_list(&p->side, 1,aacdec_info);
    App_get_ele_list(&p->back, 1,aacdec_info);
    App_get_ele_list(&p->lfe, 0,aacdec_info);
    App_get_ele_list(&p->data, 0,aacdec_info);
    App_get_ele_list(&p->coupling, 1,aacdec_info);

    App_bs_byte_align(aacdec_info);

    j = App_bs_read_bits(LEN_COMMENT_BYTES,aacdec_info);


    /*
     * The comment bytes are overwritten onto the same location, to
     * save memory.
     */

    for (i = 0; i < j; i++)
        p->comments[0] = App_bs_read_bits(LEN_BYTE,aacdec_info);
    /* null terminator for string */
    p->comments[0] = 0;

    return 0;

}



/*******************************************************************************
 *
 *   FUNCTION NAME - App_get_adif_header
 *
 *   DESCRIPTION
 *       Gets ADIF header from the input bitstream.
 *
 *   ARGUMENTS
 *         params  -  place to store the adif-header data
 *
 *   RETURN VALUE
 *         Success :  1
 *         Error   : -1
*******************************************************************************/
int App_get_adif_header(AACD_Block_Params * params,MFW_GST_AACDEC_INFO_T *aacdec_info)
{
    int             i,
                    n,
                    select_status;
    AACD_ProgConfig      *tmp_config;
    ADIF_Header     temp_adif_header;
    ADIF_Header*    p = &(temp_adif_header);

    /* adif header */
    for (i = 0; i < LEN_ADIF_ID; i++)
         p->adif_id[i] = App_bs_read_bits(LEN_BYTE,aacdec_info);
    p->adif_id[i] = 0;  /* null terminated string */

#ifdef UNIX
    /* test for id */
    if (strncmp(p->adif_id, "ADIF", 4) != 0)
        return -1;  /* bad id */
#else
    /* test for id */
    if (*((unsigned long *) p->adif_id) != *((unsigned long *) "ADIF"))
        return -1;  /* bad id */
#endif

    /* copyright string */
    if ((p->copy_id_present = App_bs_read_bits(LEN_COPYRT_PRES,aacdec_info)) == 1)
    {
        for (i = 0; i < LEN_COPYRT_ID; i++)
            p->copy_id[i] = (char)App_bs_read_bits(LEN_BYTE,aacdec_info);

        /* null terminated string */
        p->copy_id[i] = 0;
    }
    p->original_copy = App_bs_read_bits(LEN_ORIG,aacdec_info);
    p->home = App_bs_read_bits(LEN_HOME,aacdec_info);
    p->bitstream_type = App_bs_read_bits(LEN_BS_TYPE,aacdec_info);
    p->bitrate = App_bs_read_bits(LEN_BIT_RATE,aacdec_info);

    /* program config elements */
    select_status = -1;
    n = App_bs_read_bits(LEN_NUM_PCE,aacdec_info) + 1;

    tmp_config = (AACD_ProgConfig*) aacd_alloc_fast (n*sizeof(AACD_ProgConfig));

    for (i = 0; i < n; i++)
    {
        tmp_config[i].buffer_fullness =
            (p->bitstream_type == 0) ? App_bs_read_bits(LEN_ADIF_BF,aacdec_info) : 0;


        if (App_get_prog_config(&tmp_config[i],aacdec_info))
        {
            return -1;
        }

	select_status = 1;
    }

    App_bs_byte_align(aacdec_info);

    /* Fill in the AACD_Block_Params struct now */
    
    params->num_pce = n;
    params->pce     = tmp_config;
    params->BitstreamType = p->bitstream_type;
    params->BitRate       = p->bitrate;
    params->ProtectionAbsent = 0;
    return select_status;
}




/*******************************************************************************
 *
 *   FUNCTION NAME - App_get_adts_header
 *
 *   DESCRIPTION
 *       Searches and syncs to ADTS header from the input bitstream. It also
 *       gets the full ADTS header once sync is obtained.
 *
 *   ARGUMENTS
 *       params   - Place to store the header data
 *
 *   RETURN VALUE
 *       Success : 0
 *         Error : -1
*******************************************************************************/
int App_get_adts_header(AACD_Block_Params * params,MFW_GST_AACDEC_INFO_T *aacdec_info)
{
    ADTS_Header App_adts_header;
    ADTS_Header *p = &(App_adts_header);

    int bits_used = 0;
	//// bitrate support for adts header - tlsbo79743
	int start_bytes=0;
	int bits_consumed=0;
		unsigned int unSampleRate;
	const unsigned int unFrameSamples = 1024 ;
	unsigned int bufferFullness;
	const unsigned char channelConfig2NCC[8] = {0,1,2,3,4,5,6,7};
#ifdef OLD_FORMAT_ADTS_HEADER
    int emphasis;
#endif


    while (1)
   {
        /*
         * If we have used up more than maximum possible frame for finding
         * the ADTS header, then something is wrong, so exit.
         */
        if (bits_used > (LEN_BYTE * ADTS_FRAME_MAX_SIZE))
            return -1;

        /* ADTS header is assumed to be byte aligned */
        bits_used += App_bs_byte_align(aacdec_info);

#ifdef CRC_CHECK
        /* Add header bits to CRC check */
        UpdateCRCStructBegin();
#endif

        p->syncword = App_bs_read_bits ((LEN_SYNCWORD - LEN_BYTE),aacdec_info);
        bits_used += LEN_SYNCWORD - LEN_BYTE;

        /* Search for syncword */
        while (p->syncword != ((1 << LEN_SYNCWORD) - 1))
        {
            p->syncword = ((p->syncword << LEN_BYTE) | App_bs_read_bits(LEN_BYTE,aacdec_info));
            p->syncword &= (1 << LEN_SYNCWORD) - 1;
            bits_used += LEN_BYTE;
            /*
             * If we have used up more than maximum possible frame for finding
             * the ADTS header, then something is wrong, so exit.
             */
            if (bits_used > (LEN_BYTE * ADTS_FRAME_MAX_SIZE))
                return -1;
	}

          bits_consumed = bits_used -  LEN_SYNCWORD; // bitrate support for adts header - tlsbo79743
	
         p->id = App_bs_read_bits(LEN_ID,aacdec_info); 
	bits_used += LEN_ID;

	/*
	  Disabled version check in allow MPEG2/4 streams
	  0 - MPEG4
	  1 - MPEG2
	  if (!p->id)
	  {
	  continue;
	  }
	*/

        p->layer = App_bs_read_bits(LEN_LAYER,aacdec_info);
	bits_used += LEN_LAYER;
	if (p->layer != 0)
	  {
	    continue;
	  }

	p->protection_abs = App_bs_read_bits(LEN_PROTECT_ABS,aacdec_info);
	bits_used += LEN_PROTECT_ABS;

	p->profile = App_bs_read_bits(LEN_PROFILE,aacdec_info);
        bits_used += LEN_PROFILE;
	if (p->profile != 1)
	  {
            continue;
	  }

	p->sampling_freq_idx = App_bs_read_bits(LEN_SAMP_IDX,aacdec_info);
        bits_used += LEN_SAMP_IDX;
	if (p->sampling_freq_idx >= 0xc)
	  {
            continue;
	  }

	p->private_bit = App_bs_read_bits(LEN_PRIVTE_BIT,aacdec_info);
        bits_used += LEN_PRIVTE_BIT;

        //temp_channel_config = p->channel_config;
	p->channel_config = App_bs_read_bits(LEN_CHANNEL_CONF,aacdec_info);
        bits_used += LEN_CHANNEL_CONF;
        ///* Audio mode has changed, so config has to be built up again */
        //if (temp_channel_config != p->channel_config)
	//ptr->AACD_default_config = 1;
	p->original_copy = App_bs_read_bits(LEN_ORIG,aacdec_info);
        bits_used += LEN_ORIG;

	p->home = App_bs_read_bits(LEN_HOME,aacdec_info);
        bits_used += LEN_HOME;

#ifdef OLD_FORMAT_ADTS_HEADER
	params->Flush_LEN_EMPHASIS_Bits = 0;
	if (p->id == 0)
	  {
	    emphasis = App_bs_read_bits(LEN_EMPHASIS,aacdec_info);
	    bits_used += LEN_EMPHASIS;
	    params->Flush_LEN_EMPHASIS_Bits = 1;
	  }

#endif

	p->copyright_id_bit = App_bs_read_bits(LEN_COPYRT_ID_ADTS,aacdec_info);
        bits_used += LEN_COPYRT_ID_ADTS;

	p->copyright_id_start = App_bs_read_bits(LEN_COPYRT_START,aacdec_info);
        bits_used += LEN_COPYRT_START;

	p->frame_length = App_bs_read_bits(LEN_FRAME_LEN,aacdec_info);
        bits_used += LEN_FRAME_LEN;

	p->adts_buffer_fullness = App_bs_read_bits(LEN_ADTS_BUF_FULLNESS,aacdec_info);
        bits_used += LEN_ADTS_BUF_FULLNESS;

	p->num_of_rdb = App_bs_read_bits(LEN_NUM_OF_RDB,aacdec_info);
        bits_used += LEN_NUM_OF_RDB;

   /*
     Removed, constraint: num_of_rdb == 0 because we can support more than
     one raw data block in 1 adts frame
     if (p->num_of_rdb != 0)
     {
     continue;
     }
   */

        /*
         * If this point is reached, then the ADTS header has been found and
         * CRC structure can be updated.
         */
#ifdef CRC_CHECK
        /*
         * Finish adding header bits to CRC check. All bits to be CRC
         * protected.
         */
        UpdateCRCStructEnd(0);
#endif
        /*
         * Adjust the received frame length to add the bytes used up to
         * find the ADTS header.
         */
        p->frame_length += (bits_used / LEN_BYTE) - ADTS_FRAME_HEADER_SIZE;

        if (p->protection_abs == 0)
	  {
	    p->crc_check = App_bs_read_bits(LEN_CRC,aacdec_info);
            bits_used += LEN_CRC;
	  }

        /* Full header successfully obtained, so get out of the search */
        break;
   }

#ifndef OLD_FORMAT_ADTS_HEADER
    bits_used += App_bs_byte_align(aacdec_info);
#else
    if (p->id != 0) // MPEG-2 style : Emphasis field is absent
      {
	bits_used += App_bs_byte_align(aacdec_info);
      }
    else //MPEG-4 style : Emphasis field is present; cancel its effect
      {
	aacdec_info->app_params.BitsInHeader -= LEN_EMPHASIS;
      }
#endif

    /* Fill in the AACD_Block_Params struct now */

    params->num_pce = 0;
    params->ChannelConfig = p->channel_config;
    params->SamplingFreqIndex = p->sampling_freq_idx;
    params->BitstreamType = (p->adts_buffer_fullness == 0x7ff) ? 1 : 0;



    params->BitRate       = 0; /* Caution !*/

    /* The ADTS stream contains the value of buffer fullness in units
       of 32-bit words. But the application is supposed to deliver this
       in units of bits. Hence we do a left-shift */

    params->BufferFullness = (p->adts_buffer_fullness) << 5;
    params->ProtectionAbsent = p->protection_abs;
    params->CrcCheck = p->crc_check;


    //ptr->AACD_mc_info.profile = p->profile;
    //ptr->AACD_mc_info.sampling_rate_idx = p->sampling_freq_idx;
    //AACD_infoinit(&(ptr->tbl_ptr_AACD_samp_rate_info[ptr->AACD_mc_info.sampling_rate_idx]), ptr);

	///////////////////////  bitrate support for adts header - tlsbo79743 /////////////////////////////
	
	bufferFullness = p->adts_buffer_fullness*32*channelConfig2NCC[p->channel_config];

	trnsptAdjustBitrate(bits_consumed,  8*p->frame_length, bufferFullness,aacdec_info);

    return 0;

}

/*******************************************************************************
 *
 *   FUNCTION NAME - trnsptAdjustBitrate
 *
 *   DESCRIPTION
 *       In this function bitsPerFrame is calculated by finding offset.
 *
 *   ARGUMENTS
 *       offset         - Bits between start of the frame and sync word
 *       frameSize	    - length of each frame (in bits)
 *       bufferFullness	- 
 *
 *   RETURN VALUE
 *       NONE
 *         
*******************************************************************************/

void trnsptAdjustBitrate(unsigned int offset,
                         unsigned int frameSize,
                         unsigned int bufferFullness,MFW_GST_AACDEC_INFO_T * aacdec_info)
{
    if (!aacdec_info->nFramesReceived) /* the first time around */
        firstBufferFullness = bufferFullness ;
    else if (aacdec_info->nBitsReceived < (1UL<<(30-BITSPERFRAME_SCALE)))
        aacdec_info->nBitsReceived += frameSize + offset ;

    if (!offset)
    {
        lastGoodBufferFullness = bufferFullness ;
        /* stop bitrate calculation before we have to do long divisions */
        if (aacdec_info->nFramesReceived && aacdec_info->nBitsReceived < (1UL<<(30-BITSPERFRAME_SCALE)))
            aacdec_info->bitsPerFrame = ((aacdec_info->nBitsReceived + bufferFullness - firstBufferFullness) <<BITSPERFRAME_SCALE) / aacdec_info->nFramesReceived ;
        aacdec_info->nFramesReceived ++ ;
        // fprintf(stderr,"bitsPerFrame = %d (received = %d)\n",h->bitsPerFrame>>BITSPERFRAME_SCALE,h->nFramesReceived);
    }
    else
    {
        /* if offset !=0, we skipped frames. Keep bitrate estimate and instead estimate
           number of frames skipped */
        if (aacdec_info->bitsPerFrame) /* only if we have a br estimate already */
        {
            unsigned int nFramesSkipped = (((offset + bufferFullness - lastGoodBufferFullness) << BITSPERFRAME_SCALE) + (aacdec_info->bitsPerFrame>>1)) / aacdec_info->bitsPerFrame ;
            aacdec_info->nFramesReceived += nFramesSkipped ;
//            fprintf(stderr,"bitsPerFrame = %d (received = %d, skipped = %d)\n",h->bitsPerFrame>>BITSPERFRAME_SCALE,h->nFramesReceived,nFramesSkipped);
        }
    }
}

