/*
 * Copyright 2005-2007 Freescale Semiconductor, Inc. All Rights Reserved.
 */

/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*=============================================================================
                                                                               
    Module Name:            mfw_gst_aacdec.c

    General Description:    Gstreamer plugin for AAC + LC decoder
                            capable of decoding AAC (with both ADIF and ADTS 
                            header).
                            
===============================================================================
Portability: This file is ported for Linux and GStreamer.

===============================================================================
                            INCLUDE FILES
=============================================================================*/

#include <gst/gst.h>
#include <string.h>
#include "aacd_dec_interface.h"
#include "mfw_gst_aacdec.h"
#include "aac_aux.h"
/*=============================================================================
                            LOCAL CONSTANTS
=============================================================================*/
#define MULT_FACTOR 4

/* the below macros are used to calculate the 
Bitrate by parsing the ADTS header */
#define SAMPLING_FREQ_IDX_MASk  0x3c
#define BITSPERFRAME_MASK 0x3ffe000
#define ADTS_HEADER_LENGTH 7 

/*=============================================================================
                LOCAL TYPEDEFS (STRUCTURES, UNIONS, ENUMS)
=============================================================================*/

static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE("sink",
								   GST_PAD_SINK,
								   GST_PAD_ALWAYS,
								   GST_STATIC_CAPS
								   ("audio/mpeg, "
								    "mpegversion	= (gint){2, 4}, "
								    "bitrate = [8000 ,	384000],"
								    "channels =	(gint) [	1, 2 ]"));

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE("src",
								  GST_PAD_SRC,
								  GST_PAD_ALWAYS,
								  GST_STATIC_CAPS("audio/x-raw-int, \
                                    " "endianness	= (gint)	" G_STRINGIFY(G_BYTE_ORDER) ",\
                                    " "signed	= (boolean)	true, \
                                    " "width = (gint) 16,\
                                    " "depth = (gint) 16, \
                                    " "rate =	(gint) {8000, 11025, 12000,	16000,\
                                    22050, 24000, 32000,44100, 48000 },	\
                                    " "channels =	(gint) [	1, 2 ]"));


/*=============================================================================
                                LOCAL MACROS
=============================================================================*/
#define LONG_BOUNDARY 4
#define	GST_CAT_DEFAULT    mfw_gst_aacdec_debug

/*=============================================================================
                        STATIC FUNCTION PROTOTYPES
=============================================================================*/
GST_DEBUG_CATEGORY_STATIC(mfw_gst_aacdec_debug);
static void mfw_gst_aacdec_class_init(MFW_GST_AACDEC_CLASS_T *
					  klass);
static void mfw_gst_aacdec_base_init(MFW_GST_AACDEC_CLASS_T *
					 klass);
static void mfw_gst_aacdec_init(MFW_GST_AACDEC_INFO_T *
				    aacdec_info);
static void mfw_gst_aacdec_set_property(GObject * object,
					    guint prop_id,
					    const GValue * value,
					    GParamSpec * pspec);
static void mfw_gst_aacdec_get_property(GObject * object,
					    guint prop_id, GValue * value,
					    GParamSpec * pspec);
static gboolean mfw_gst_aacdec_set_caps(GstPad * pad, GstCaps * caps);
static GstFlowReturn mfw_gst_aacdec_chain(GstPad * pad,
					      GstBuffer * buf);
static gboolean mfw_gst_aacdec_sink_event(GstPad *, GstEvent *);
static gboolean mfw_gst_aacdec_src_event(GstPad *, GstEvent *);
static gboolean plugin_init(GstPlugin * plugin);
static gboolean mfw_gst_aacdec_seek(MFW_GST_AACDEC_INFO_T *,
					GstPad *, GstEvent *);
static gboolean mfw_gst_aacdec_convert_src(GstPad * pad,
					       GstFormat src_format,
					       gint64 src_value,
					       GstFormat * dest_format,
					       gint64 * dest_value);
static gboolean mfw_gst_aacdec_convert_sink(GstPad * pad,
						GstFormat src_format,
						gint64 src_value,
						GstFormat * dest_format,
						gint64 * dest_value);
static gboolean mfw_gst_aacdec_src_query(GstPad *, GstQuery *);
static const GstQueryType *mfw_gst_aacdec_get_query_types(GstPad *
							      pad);
static gboolean mfw_gst_aacdec_mem_flush(MFW_GST_AACDEC_INFO_T *);

/*=============================================================================
                            STATIC VARIABLES
=============================================================================*/
static GstElementClass *parent_class_aac = NULL;
static MFW_GST_AACDEC_INFO_T *aacdec_global_ptr = NULL;

/*=============================================================================
                            LOCAL FUNCTIONS
=============================================================================*/



/*=============================================================================
FUNCTION: mfw_gst_aacdec_set_property

DESCRIPTION: sets the property of the element

ARGUMENTS PASSED:
        object     - pointer to the elements object
        prop_id    - ID of the property;
        value      - value of the property set by the application
        pspec      - pointer to the attributes of the property

RETURN VALUE:
        None

PRE-CONDITIONS:
        None

POST-CONDITIONS:


IMPORTANT NOTES:
        None
=============================================================================*/
static void
mfw_gst_aacdec_set_property(GObject * object, guint prop_id,
				const GValue * value, GParamSpec * pspec)
{

    GST_DEBUG(" in mfw_gst_aacdec_set_property routine \n");
    GST_DEBUG(" out of mfw_gst_aacdec_set_property routine \n");
}

/*=============================================================================
FUNCTION: mfw_gst_aacdec_set_property

DESCRIPTION: gets the property of the element

ARGUMENTS PASSED:
        object     - pointer to the elements object
        prop_id    - ID of the property;
        value      - value of the property got from the application
        pspec      - pointer to the attributes of the property

RETURN VALUE:
        None

PRE-CONDITIONS:
        None

POST-CONDITIONS:


IMPORTANT NOTES:
        None
=============================================================================*/
static void
mfw_gst_aacdec_get_property(GObject * object, guint prop_id,
				GValue * value, GParamSpec * pspec)
{
    GST_DEBUG(" in mfw_gst_aacdec_get_property routine \n");
    GST_DEBUG(" out of mfw_gst_aacdec_get_property routine \n");

}


/***************************************************************************
*
*   FUNCTION NAME - alloc_fast
*
*   DESCRIPTION
*          This function simulates to allocate memory in the internal
*          memory. This function when used by the application should
*          ensure that the memory address returned in always aligned
*          to long boundry.
*
*   ARGUMENTS
*       size              - Size of the memory requested.
*
*   RETURN VALUE
*       base address of the memory chunck allocated.
*
***************************************************************************/
 void *aacd_alloc_fast(gint size)
{
    void *ptr = NULL;
    ptr = (void *) g_malloc(size + 4);
    ptr =
	(void *) (((long) ptr + (long) (LONG_BOUNDARY - 1)) &
		  (long) (~(LONG_BOUNDARY - 1)));
    return ptr;

}

/***************************************************************************
*
*   FUNCTION NAME - app_swap_buffers_aac_dec
*
*   DESCRIPTION
*       This function is a callback by instance of AACD decoder requesting
*       the application for a new input buffer. It returns the used buffer
*       pointer to the application and expects a new buffer pointer and
*       buffer length in return. If buffer is not available it returns
*       a NULL pointer with zero length to the AACD decoder instance. This
*       is an example code.
*
*       The variable bytes_supplied keeps track of the total number
*       of bytes used by the decoder, since the last call to the decoder.
*
*       Also see, update_bitstream_status() function below.
*
*   ARGUMENTS
*       new_buf_ptr       - Pointer to pointer of used buffer
*       new_buf_len       - Pointer to the length of the input buffer var
*       instance_id       - AACD decoder Instance ID
*
*   RETURN VALUE
*          0              - If buffer allocation is successfull.
*         -1              - Indicates 'End of Bitstream'.
*
***************************************************************************/

AACD_INT8 app_swap_buffers_aac_dec (AACD_UCHAR **new_buf_ptr,
                               AACD_UINT32 *new_buf_len,
                               void *global_struct)
{

    gint len;
    if (!aacdec_global_ptr->inbuffer2) {
	*new_buf_ptr = NULL;
	*new_buf_len = 0;
	return (AACD_INT8) - 1;
    }
    *new_buf_ptr =
	(AACD_UCHAR *) GST_BUFFER_DATA(aacdec_global_ptr->inbuffer2);
    *new_buf_len = GST_BUFFER_SIZE(aacdec_global_ptr->inbuffer2);

    return (0);
}

/***************************************************************************
*
*   FUNCTION NAME - alloc_slow
*
*   DESCRIPTION
*          This function simulates to allocate memory in the internal
*          memory. This function when used by the application should
*          ensure that the memory address returned in always aligned
*          to long boundry.
*
*   ARGUMENTS
*       size              - Size of the memory requested.
*
*   RETURN VALUE
*       base address of the memory chunck allocated.
*
***************************************************************************/


static void *aacd_alloc_slow(gint size)
{
    void *ptr = NULL;
    ptr = (void *) g_malloc(size);
    ptr =
	(void *) (((long) ptr + (long) LONG_BOUNDARY - 1) &
		  (long) (~(LONG_BOUNDARY - 1)));
    return ptr;
}

/***************************************************************************
*
*   FUNCTION NAME - aacd_free
*
*   DESCRIPTION     This function frees the memory allocated previously
*   ARGUMENTS
*       mem       - memory address to be freed
*
*   RETURN VALUE
*       None
*
***************************************************************************/

static void aacd_free(void *mem)
{
    g_free(mem);
    return;
}


/***************************************************************************
*
*   FUNCTION NAME - mfw_gst_interleave_samples
*
*   DESCRIPTION
*                   This function interleaves the decoded left and right 
*                   channel output PCM samples
*   ARGUMENTS
*       data_ch0    - pointer to the decoded left channel output PCM samples
*       data_ch1    - pointer to the decoded right channel output PCM samples
*       data_out    - pointer to store the interleaved output
*       frameSize   - output frame size
*       channels    - number of output channels
*
*   RETURN VALUE
*       None
*
***************************************************************************/
static void mfw_gst_interleave_samples(AACD_OutputFmtType * data_ch0,
				       AACD_OutputFmtType * data_ch1,
				       short * data_out,
				       gint frameSize, gint channels)
{
    gint i;
    AACD_OutputFmtType tmp;

    for (i = 0; i < frameSize; i++) {
	*data_out++ = *data_ch0++;

	if (channels == 2)
	    *data_out++ = *data_ch1++;
    }

}

/***************************************************************************
*
*   FUNCTION NAME - AACD_writeout
*
*   DESCRIPTION
*                   This function processes the output depending on the 
*                   number of channels and the stream type
*   ARGUMENTS
*       dec_info    - pointer to decoder output info structure
*       data        - output data
*       mip         - pointer to multi-channel information
*       outbuff     - output buffer
*
*   RETURN VALUE
*       0           - no error in formatting the output
*
***************************************************************************/
static gint AACD_writeout(AACD_Decoder_info * dec_info,AACD_Decoder_Config *dec_config,
			  AACD_OutputFmtType data[][AAC_FRAME_SIZE],guint8 * outbuff)
{

    
    gint i,num_chans,j;
    gint localVar[Chans];
    
    num_chans = Chans;
    

    if (dec_info->aacd_num_channels == 1) {
	mfw_gst_interleave_samples(&data[0][0], &data[0][0],
				   (short *) outbuff,
				   dec_info->aacd_len, 2);
    } 

    else if (dec_info->aacd_len != AAC_FRAME_SIZE) 
    {
        
        
    /* this is possible for AAC LC test vectors
		 	* This check is done so that parametric stereo
            * data can be iWritten to file properly
            * in case of aac vectors stereo o/p will be present in
            * data[0] and data[1] whereas in PS it will be in
            * data[0] and data[1]
        */
        j = 0;
        for (i = 0; i < num_chans; i++) 
        {
            if (!(*(dec_config->ch_info[i].present)))		
                continue;
            localVar[j] = i;
            j++;
        }
        
        mfw_gst_interleave_samples(&data[localVar[1]][0],&data[localVar[2]][0],
            (short*) outbuff,dec_info->aacd_len, 2);
    }
    else {
	mfw_gst_interleave_samples(&data[1][0], &data[2][0],
				   (short *) outbuff,
				   dec_info->aacd_len, 2);
    } 

    return 0;

}

/***************************************************************************
*
*   FUNCTION NAME - mfw_gst_aacdec_data
*
*   DESCRIPTION
*                   This function decodes data in the input buffer and 
*                   pushes the decode pcm output to the next element in the 
*                   pipeline
*   ARGUMENTS
*       aacdec_info    - pointer to the plugin context
*       inbuffsize         - pointer to the input buffer size
*
*   RETURN VALUE
*       TRUE               - decoding is succesful
*       FALSE              - error in decoding
***************************************************************************/

static gboolean
mfw_gst_aacdec_data(MFW_GST_AACDEC_INFO_T * aacdec_info,
			gint * inbuffsize)
{

    AACD_RET_TYPE rc;
    gint rec_no = 0;
    GstCaps *src_caps = NULL;
    GstCaps *caps = NULL;
    GstBuffer *outbuffer = NULL;
    guint8 *inbuffer = NULL;
    GstFlowReturn res = GST_FLOW_ERROR;
    guint64 time_duration = 0;
    AACD_Decoder_Config *dec_config = NULL;
    AACD_Decoder_info dec_info;
    AACD_OutputFmtType outbuf[Chans][AAC_FRAME_SIZE];
    dec_config = aacdec_info->app_params.dec_config;
    GstBuffer *residue=NULL;

    /* No buf data need to be decoded, just return */
    if (*inbuffsize == 0)
        return TRUE;

    do {

	if (aacdec_global_ptr == NULL) {

	    return TRUE;
	}

	/* decoding of ADTS header happens here */
	if (aacdec_info->app_params.App_adts_header_present) {

        inbuffer = GST_BUFFER_DATA(aacdec_info->inbuffer1) +
		   GST_BUFFER_OFFSET(aacdec_info->inbuffer1);
	App_bs_readinit((gchar *) inbuffer, *inbuffsize,
			    aacdec_info);
	aacdec_info->app_params.BitsInHeader = 0;

        if(*inbuffsize < ADTS_HEADER_LENGTH)
        {
            GST_DEBUG("Input less than ADTS_HEADER_LENGTH\n");
            if(aacdec_info->inbuffer2 == NULL)
            {
                /* no buffer is there to join in case of EOS */
                return TRUE;
            }
            else
            {
                residue = gst_buffer_new_and_alloc(*inbuffsize);
                memcpy(GST_BUFFER_DATA(residue),inbuffer,*inbuffsize);
                aacdec_info->inbuffer2 = 
                    gst_buffer_join(residue,aacdec_info->inbuffer2);
                gst_buffer_unref(aacdec_info->inbuffer1);
                aacdec_info->inbuffer1=NULL;
                GST_BUFFER_OFFSET(aacdec_info->inbuffer2)=0;
                *inbuffsize=0;

                if(aacdec_info->corrupt_bs)
                {
                    aacdec_info->corrupt_bs = FALSE;
                    if (GST_CLOCK_TIME_IS_VALID(aacdec_info->buffer_time))
                        aacdec_info->time_offset = aacdec_info->buffer_time;
                }
                return TRUE;
            }

        }

        if(App_get_adts_header(&(aacdec_info->app_params.BlockParams),
            aacdec_info) != 0 )
        {
                GST_DEBUG("No sync found in this buffer\n");
                gst_buffer_unref(aacdec_info->inbuffer1);
                aacdec_info->inbuffer1=NULL;
                *inbuffsize=0;
                if (GST_CLOCK_TIME_IS_VALID(aacdec_info->buffer_time))
                    aacdec_info->time_offset = aacdec_info->buffer_time;
                aacdec_info->corrupt_bs = FALSE;
                return TRUE;
        }


	    dec_config->params =
		&(aacdec_info->app_params.BlockParams);
	    GST_BUFFER_OFFSET(aacdec_info->inbuffer1) +=
		(aacdec_info->app_params.BitsInHeader / 8);
	    *inbuffsize =
		*inbuffsize -
		(aacdec_info->app_params.BitsInHeader / 8);
	}

	inbuffer = GST_BUFFER_DATA(aacdec_info->inbuffer1) +
	           GST_BUFFER_OFFSET(aacdec_info->inbuffer1);
        GST_DEBUG(" Begin to decode aac frame");
	/* the decoder decodes the encoded data in the input buffer and outputs 
	   a frame of PCM samples */
        rc = aacd_decode_frame(dec_config,&dec_info,outbuf, inbuffer, *inbuffsize);
	GST_DEBUG(" return val of decdoe = %d\n", rc);
        if ((rc != AACD_ERROR_NO_ERROR && rc != AACD_ERROR_EOF)) {
            if(rc < AACD_ERROR_INIT && aacdec_info->app_params.App_adts_header_present)
            {
                dec_info.aacd_sampling_frequency = aacdec_info->sampling_freq;
                dec_info.aacd_num_channels = aacdec_info->number_of_channels;
                dec_info.aacd_len = AAC_FRAME_SIZE; 
                aacdec_info->corrupt_bs = TRUE;
                memset(outbuf, 0, AAC_FRAME_SIZE * Chans * sizeof(AACD_OutputFmtType));
            }
            else
            {
                aacdec_info->flow_error = TRUE;
                return FALSE;
            }

            GST_ERROR("Error in decoding the frame error is %d\n",rc);
            GST_DEBUG("inbuffsize = %d\n",*inbuffsize);
	}


	aacdec_info->sampling_freq = dec_info.aacd_sampling_frequency;
	aacdec_info->number_of_channels = dec_info.aacd_num_channels;
    

	GST_BUFFER_OFFSET(aacdec_info->inbuffer1) +=
	    (dec_info.BitsInBlock / 8);
	*inbuffsize -= (dec_info.BitsInBlock / 8);

    
        if(*(dec_config->AACD_bno) <= 2)
        continue;

    
        if (dec_info.aacd_len != 0 && dec_info.aacd_num_channels != 0) {
    
        
    /* capabailites of the src pad are set in accordance with the next osssink
        element in the pipeline */
        if (!aacdec_info->caps_set) {
            aacdec_info->caps_set = TRUE;
            caps = gst_caps_new_simple("audio/x-raw-int",
                "endianness", G_TYPE_INT,
                G_BYTE_ORDER, "signed",
                G_TYPE_BOOLEAN, TRUE, "width",
                G_TYPE_INT, 16, "depth", G_TYPE_INT,
                16, "rate", G_TYPE_INT,
                dec_info.aacd_sampling_frequency,
                "channels", G_TYPE_INT, 2, NULL);
            gst_pad_set_caps(aacdec_info->srcpad, caps);
            gst_caps_unref(caps);

            if(aacdec_info->app_params.App_adif_header_present)
            {
                aacdec_info->bit_rate=dec_info.aacd_bit_rate;
                GST_DEBUG(" ADIF FILE Bit Rate =  %d \n",
                    aacdec_info->bit_rate );
            }
          

        }
        
        src_caps = GST_PAD_CAPS(aacdec_info->srcpad);
        
        /* multiplication factor is obtained as a multiple of Bytes per
	       sample and number of channels */
       
        res =
            gst_pad_alloc_buffer_and_set_caps(aacdec_info->srcpad,
            0,
            MULT_FACTOR  * 
            dec_info.aacd_len,
            src_caps, &outbuffer);
        
        if (res != GST_FLOW_OK) {
            GST_DEBUG("Error in allocating output buffer");
            return FALSE;
        }
        /* output samples are interleaved based on the number of channels
	       and the type of stream */
      
        if (AACD_writeout(&dec_info, dec_config, outbuf, GST_BUFFER_DATA(outbuffer)) != 0) {
            GST_ERROR("Application File I/O Error\n");
            return FALSE;
        }
       
        GST_BUFFER_SIZE(outbuffer) = MULT_FACTOR * dec_info.aacd_len;
        time_duration = gst_util_uint64_scale_int(dec_info.aacd_len, GST_SECOND, dec_info.aacd_sampling_frequency);
        
        /* The timestamp in nanoseconds     of the data     in the buffer. */
        GST_BUFFER_TIMESTAMP(outbuffer) = aacdec_info->time_offset;
        
        /* The duration     in nanoseconds of the data in the buffer */
        GST_BUFFER_DURATION(outbuffer) = time_duration;
        /* The offset in the source file of the     beginning of this buffer */
        GST_BUFFER_OFFSET(outbuffer) = 0;
        /*record next timestamp */       
        aacdec_info->time_offset += time_duration; 
        
        /* Output PCM samples are pushed on to the next element in the pipeline */
        res = gst_pad_push(aacdec_info->srcpad, outbuffer);
        if (res != GST_FLOW_OK) {
            GST_ERROR(" not able to push the data \n");
            return FALSE;
        }
    } else {
	    aacdec_info->init_done = FALSE;
	    return TRUE;
	}


    } while (*inbuffsize > 0);

    if(aacdec_info->corrupt_bs)
    {
        aacdec_info->corrupt_bs = FALSE;
        if (GST_CLOCK_TIME_IS_VALID(aacdec_info->buffer_time))
            aacdec_info->time_offset = aacdec_info->buffer_time;
    }

    return TRUE;

}

/***************************************************************************
*
*   FUNCTION NAME - mfw_gst_aacdec_calc_average_bitrate
*
*   DESCRIPTION
*                   This function calulates the average bitrate by 
                    parsing the input stream.
*   ARGUMENTS
*       aacdec_info    - pointer to the plugin context
*
*   RETURN VALUE
*       TRUE               - execution succesful
*       FALSE              - error in execution
***************************************************************************/

static gint mfw_gst_aacdec_calc_average_bitrate(MFW_GST_AACDEC_INFO_T *
					aacdec_info)
{
    GstPad *pad = NULL;
    GstPad *peer_pad = NULL;
    GstFormat fmt = GST_FORMAT_BYTES;
    pad = aacdec_info->sinkpad;
    GstBuffer *pullbuffer = NULL;
    GstFlowReturn ret = GST_FLOW_OK;
    guint pullsize = 0;
    guint64  offset = 0;
    gfloat  avg_bitrate=0;
    guint64 totalduration = 0;
    guint64 bitrate=0;
    guint  frames=0;
    guint8 *inbuffer=NULL;
    AACD_Decoder_info dec_info;
    guint temp = 0;
    gint FileType=0;
    GstClockTime file_duration;
    if (gst_pad_check_pull_range(pad)) 
    {
        if (gst_pad_activate_pull(GST_PAD_PEER(pad), TRUE)) 
        {
            peer_pad = gst_pad_get_peer(aacdec_info->sinkpad);
            gst_pad_query_duration(peer_pad, &fmt, &totalduration);
            gst_object_unref(GST_OBJECT(peer_pad));
            pullsize = 4;
            ret = gst_pad_pull_range(pad, offset, pullsize,
				       &pullbuffer);
            App_bs_readinit((gchar *) GST_BUFFER_DATA(pullbuffer), 
                pullsize, aacdec_info);
            FileType = App_bs_look_bits(32, aacdec_info);
            if (App_FindFileType(FileType, aacdec_info) != 0) {
                GST_ERROR("InputFile is not AAC\n");
                return -1;
            }
            if(aacdec_info->app_params.App_adif_header_present==TRUE)
            {
                gst_pad_activate_push(GST_PAD_PEER(pad), TRUE);
                return 0;
            }
            pullsize = ADTS_HEADER_LENGTH;
            while(offset<totalduration)
            {
                ret = gst_pad_pull_range(pad, offset, pullsize,
				       &pullbuffer);

                /* The sampling frequency index is of length 4 bits
                after 18 bits (i.e bit3 to bit6 of the 3rd byte )
                in the ADTS header */

                inbuffer = GST_BUFFER_DATA(pullbuffer);
                temp = *(inbuffer + 2);
                temp = temp & SAMPLING_FREQ_IDX_MASk;
                temp = temp >> 2;
                aacdec_info->SampFreqIdx = temp;

                /* The Frame Length is of length 13 bits
                after 30 bits (i.e bit7 to bit8 of the 4th byte + 
                5th byte complete + bit1 to bit3 of the 6th byte)
                in the ADTS header */

                temp = ((*(inbuffer + 3)) << 24) | ((*(inbuffer + 4)) << 16) 
                        | ((*(inbuffer + 5)) << 8) | (*(inbuffer + 6));
                temp = temp & BITSPERFRAME_MASK;
                temp = temp >> 13;
                /* If frame length is 0, should quit the while cycle */
                if (temp == 0) {
                    GST_DEBUG(" frame length is 0! ");
                    break;
                }
                aacdec_info->bitsPerFrame = temp*8;

                ADTSBitrate(&dec_info, aacdec_info);
                offset += temp;
                frames++;
                bitrate += dec_info.aacd_bit_rate;
           }
        }
        gst_pad_activate_push(GST_PAD_PEER(pad), TRUE);
        avg_bitrate = (gfloat)bitrate/frames;
        aacdec_info->bit_rate =
                ((avg_bitrate + 500) / 1000) * 1000;
        GST_DEBUG("avg_bitrate=%d \n",aacdec_info->bit_rate);
        file_duration = gst_util_uint64_scale(totalduration,GST_SECOND * 8,aacdec_info->bit_rate);
        GST_DEBUG(" file_duration = %" GST_TIME_FORMAT, GST_TIME_ARGS(file_duration));

    }
    return 0;
}

/***************************************************************************
*
*   FUNCTION NAME - mfw_gst_aacdec_memclean
*
*   DESCRIPTION
*                   This function frees all the memory allocated for the 
*                   plugin;
*   ARGUMENTS
*       aacdec_info    - pointer to the plugin context
*
*   RETURN VALUE
*       None
*
***************************************************************************/
static void mfw_gst_aacdec_memclean(MFW_GST_AACDEC_INFO_T *
					aacdec_info)
{

    AACD_Decoder_Config *dec_config = NULL;
    AACD_Mem_Alloc_Info_Sub *mem = NULL;
    gint nr = 0;
    gint rec_no = 0;
    GST_DEBUG("in mfw_gst_aacdec_memclean \n");
    dec_config = aacdec_info->app_params.dec_config;
    if (dec_config != NULL) {
	nr = dec_config->aacd_mem_info.aacd_num_reqs;
	for (rec_no = 0; rec_no < nr; rec_no++) {
	    mem = &(dec_config->aacd_mem_info.mem_info_sub[rec_no]);

	    if (mem->app_base_ptr) {
		aacd_free(mem->app_base_ptr);
		mem->app_base_ptr = 0;
	    }

	}
	aacd_free(dec_config);
    }
    GST_DEBUG("out of mfw_gst_aacdec_memclean \n");
}

/*=============================================================================
FUNCTION: mfw_gst_aacdec_chain

DESCRIPTION: Initializing the decoder and calling the actual decoding function

ARGUMENTS PASSED:
        pad     - pointer to pad
        buffer  - pointer to received buffer

RETURN VALUE:
        GST_FLOW_OK		- Frame decoded successfully
		GST_FLOW_ERROR	- Failure

PRE-CONDITIONS:
        None

POST-CONDITIONS:


IMPORTANT NOTES:
        None
=============================================================================*/
static GstFlowReturn
mfw_gst_aacdec_chain(GstPad * pad, GstBuffer * buf)
{


    MFW_GST_AACDEC_INFO_T *aacdec_info;
    AACD_RET_TYPE rc = 0;
    GstCaps *src_caps = NULL, *caps = NULL;
    GstBuffer *outbuffer = NULL;
    GstFlowReturn result = GST_FLOW_OK;
    guint8 *inbuffer;
    gint inbuffsize;
    gint FileType;
    gboolean ret;
    guint64 time_duration = 0;
    AACD_Decoder_Config *dec_config = NULL;
    gint i = 0;
    aacdec_info = MFW_GST_AACDEC(GST_OBJECT_PARENT(pad));
    if (aacdec_global_ptr == NULL)
	aacdec_global_ptr = aacdec_info;
    else {
	if (aacdec_global_ptr != aacdec_info) {
	    GstFlowReturn res;
	    res = gst_pad_push(aacdec_info->srcpad, buf);
	    if (res != GST_FLOW_OK) {
		GST_ERROR("aac dec:could not push onto \
                    next element %d\n", res);
	    }
	    return GST_FLOW_OK;
	}
    }

    aacdec_info->buffer_time = GST_BUFFER_TIMESTAMP(buf);

    if (!aacdec_info->init_done) {
	if (GST_CLOCK_TIME_IS_VALID(GST_BUFFER_TIMESTAMP(buf))) {
		aacdec_info->time_offset = GST_BUFFER_TIMESTAMP(buf);
	}        
	dec_config = aacdec_info->app_params.dec_config;
	aacdec_info->inbuffer1 = buf;
        GST_BUFFER_OFFSET(aacdec_info->inbuffer1)=0;
	inbuffer = GST_BUFFER_DATA(aacdec_info->inbuffer1);
	inbuffsize = GST_BUFFER_SIZE(aacdec_info->inbuffer1);

	/* searching for sync word so that can give correct 
        data to the decode library */
	if (aacdec_info->app_params.App_adts_header_present) {
        while(1)
        {
            if((inbuffer[i]==0xFF) && ((inbuffer[i+1]==0xF9) 
                || (inbuffer[i+1]==0xF1)))
            {
                break;
            }
            i++;
        }
	    inbuffer = inbuffer + i;
	}
	inbuffsize = inbuffsize - i;
	GST_BUFFER_OFFSET(aacdec_info->inbuffer1) = 0;


    
	App_bs_readinit((gchar *) inbuffer, inbuffsize, aacdec_info);
	FileType = App_bs_look_bits(32, aacdec_info);

    /* determine whether input stream has ADTS or ADIF header */
	if (App_FindFileType(FileType, aacdec_info) != 0) {
	    GST_ERROR("InputFile is not AAC\n");
	    return GST_FLOW_ERROR;
	}


	/* decoding of ADIF header happens here */
	if (aacdec_info->app_params.App_adif_header_present) {
	    GST_DEBUG("the input stream is of ADIF format \n");

            /*
             * ENGR63488: Don't support seek for ADIF streams
             */
	    aacdec_info->seek_flag = FALSE;

	    App_bs_readinit((gchar *) inbuffer, inbuffsize,
			    aacdec_info);
	    aacdec_info->app_params.BitsInHeader = 0;
        if(App_get_adif_header(&(aacdec_info->app_params.BlockParams),
            aacdec_info) < 0)
        {
            GError *error = NULL;
            GQuark domain;
            domain = g_quark_from_string("mfw_aacdecoder");
            error = g_error_new(domain, 10, "fatal error");
            gst_element_post_message(GST_ELEMENT(aacdec_info),
                gst_message_new_error(GST_OBJECT
                (aacdec_info),
                error,
                "Error ehile parsing the ADIF header in the "
                " AAC LC decoder plug-in"));
        }

	    dec_config->params =
		&(aacdec_info->app_params.BlockParams);
	    GST_BUFFER_OFFSET(aacdec_info->inbuffer1) +=
		(aacdec_info->app_params.BitsInHeader / 8);
	    GST_BUFFER_SIZE(aacdec_info->inbuffer1) -=
		(aacdec_info->app_params.BitsInHeader / 8);


	}

	aacdec_info->init_done = TRUE;
    return GST_FLOW_OK;
    }
    aacdec_info->inbuffer2 = buf;
    GST_BUFFER_OFFSET(aacdec_info->inbuffer2) = 0;
    inbuffsize = GST_BUFFER_SIZE(aacdec_info->inbuffer1);
    ret = mfw_gst_aacdec_data(aacdec_info, &inbuffsize);
    if (aacdec_info->inbuffer1) {
        
        gst_buffer_unref(aacdec_info->inbuffer1);
        aacdec_info->inbuffer1 = NULL;
    }
    aacdec_info->inbuffer1 = aacdec_info->inbuffer2;
    GST_BUFFER_OFFSET(aacdec_info->inbuffer1) -= (inbuffsize);
    GST_BUFFER_SIZE(aacdec_info->inbuffer1) += (inbuffsize);
    if (aacdec_info->flow_error) {
        GST_ERROR(" flow error !");
        GError *error = NULL;
        GQuark domain;
        domain = g_quark_from_string("mfw_aacdecoder");
        error = g_error_new(domain, 10, "fatal error");
        gst_element_post_message(GST_ELEMENT(aacdec_info),
                gst_message_new_error(GST_OBJECT
                (aacdec_info),
                error,
                "Flow error because the parsing frame length is 0 "
                " AAC decoder plug-in"));
        return GST_FLOW_ERROR;
    }
    GST_DEBUG(" out of mfw_gst_aacdec_chain routine \n");
    return GST_FLOW_OK;

}

/*=============================================================================
FUNCTION:   mfw_gst_aacdec_change_state

DESCRIPTION: this function keeps track of different states of pipeline.

ARGUMENTS PASSED:
        element     -   pointer to element
        transition  -   state of the pipeline

RETURN VALUE:
        GST_STATE_CHANGE_FAILURE    - the state change failed
        GST_STATE_CHANGE_SUCCESS    - the state change succeeded
        GST_STATE_CHANGE_ASYNC      - the state change will happen
                                        asynchronously
        GST_STATE_CHANGE_NO_PREROLL - the state change cannot be prerolled

PRE-CONDITIONS:
        None

POST-CONDITIONS:
        None

IMPORTANT NOTES:
        None
=============================================================================*/
static GstStateChangeReturn
mfw_gst_aacdec_change_state(GstElement * element,
				GstStateChange transition)
{


    MFW_GST_AACDEC_INFO_T *aacdec_info;
    GstStateChangeReturn ret = GST_STATE_CHANGE_SUCCESS;
    gint rec_no = 0;
    gint nr=0,retval=0;
    AACD_Decoder_Config *dec_config = NULL;
    AACD_Mem_Alloc_Info_Sub *mem;
    AACD_RET_TYPE rc = 0;
    gboolean res;
    aacdec_info = MFW_GST_AACDEC(element);


    GST_DEBUG(" in mfw_gst_aacdec_change_state routine \n");
    switch (transition) {
    case GST_STATE_CHANGE_NULL_TO_READY:


	aacdec_info->caps_set = FALSE;
	aacdec_info->init_done = FALSE;
	aacdec_info->eos = FALSE;
        aacdec_info->flow_error = FALSE;
	aacdec_info->time_offset = 0;
	aacdec_info->inbuffer1 = NULL;
	aacdec_info->inbuffer2 = NULL;
    aacdec_info->corrupt_bs = FALSE;

	memset(&aacdec_info->app_params, 0,
	       sizeof(AACD_App_params *));

	/* allocate memory for config structure */
	dec_config = (AACD_Decoder_Config *)
	    aacd_alloc_fast(sizeof(AACD_Decoder_Config));
	aacdec_info->app_params.dec_config = dec_config;
	if (dec_config == NULL) {
	    GST_ERROR("error in allocation of decoder config structure");
	    return GST_STATE_CHANGE_FAILURE;
	}



	/* call query mem function to know mem requirement of library */

	if (aacd_query_dec_mem(dec_config) != AACD_ERROR_NO_ERROR) {
	    GST_ERROR
		("Failed to get the memory configuration for the decoder\n");
	    return GST_STATE_CHANGE_FAILURE;
	}

	/* Number of memory chunk requests by the decoder */
	nr = dec_config->aacd_mem_info.aacd_num_reqs;

	 for(rec_no = 0; rec_no < nr; rec_no++)
     {
         mem = &(dec_config->aacd_mem_info.mem_info_sub[rec_no]);
         
         if (mem->aacd_type == AACD_FAST_MEMORY)
         {
             mem->app_base_ptr = aacd_alloc_fast (mem->aacd_size);
             if (mem->app_base_ptr == NULL)
                 return GST_STATE_CHANGE_FAILURE;
         }
         else
         {
             mem->app_base_ptr = aacd_alloc_slow (mem->aacd_size);
             if (mem->app_base_ptr == NULL)
                 return GST_STATE_CHANGE_FAILURE;
         }
         memset(dec_config->aacd_mem_info.mem_info_sub[rec_no].app_base_ptr,
             0, dec_config->aacd_mem_info.mem_info_sub[rec_no].aacd_size); 
         
     }

	aacdec_info->app_params.BitsInHeader = 0;
	aacdec_info->app_params.App_adif_header_present = FALSE;
	aacdec_info->app_params.App_adts_header_present = FALSE;

	/* register the call-back function in the decoder context */
	dec_config->app_swap_buf = app_swap_buffers_aac_dec;

#ifndef OUTPUT_24BITS
            dec_config->num_pcm_bits     = AACD_16_BIT_OUTPUT;
#else
            dec_config->num_pcm_bits     = AACD_24_BIT_OUTPUT;
#endif  /*OUTPUT_24BITS*/				/*OUTPUT_24BITS */


	rc = aacd_decoder_init(dec_config);
	if (rc != AACD_ERROR_NO_ERROR) {
	    GST_ERROR("Error in initializing the decoder");
	    return GST_STATE_CHANGE_FAILURE;
	}
	break;

    case GST_STATE_CHANGE_READY_TO_PAUSED:
	{

	    aacdec_info->bitsPerFrame = 0;
	    aacdec_info->bit_rate = 0;
	    aacdec_info->nFramesReceived = 0;
	    aacdec_info->bitstream_count = 0;
	    aacdec_info->bitstream_buf_index = 0;
	    aacdec_info->in_buf_done = 0;
	    aacdec_info->nBitsReceived = 0;
	    aacdec_info->total_time = 0;
	    aacdec_info->seek_flag = FALSE;

        retval = mfw_gst_aacdec_calc_average_bitrate(aacdec_info);
        if(retval != 0)
        {
            GST_ERROR("error in Calculating the average Bitrate\n");
            return GST_STATE_CHANGE_FAILURE;

        }
	    if (gst_pad_check_pull_range(aacdec_info->sinkpad)) {
		gst_pad_set_query_function(aacdec_info->srcpad,
			       GST_DEBUG_FUNCPTR
			       (mfw_gst_aacdec_src_query));
		aacdec_info->seek_flag = TRUE;
	    }
	    aacdec_info->total_frames = 0;
      aacdec_info->time_offset = 0;
	}
	break;

    case GST_STATE_CHANGE_PAUSED_TO_PLAYING:
	break;
    default:
	break;
    }

    ret = parent_class_aac->change_state(element, transition);

    switch (transition) {
    case GST_STATE_CHANGE_PLAYING_TO_PAUSED:
	break;
    case GST_STATE_CHANGE_PAUSED_TO_READY:
	GST_DEBUG("GST_STATE_CHANGE_PAUSED_TO_READY \n");
        /* Following is memory leak fix */   
        if ((aacdec_info->inbuffer1) && (aacdec_info->inbuffer1 != aacdec_info->inbuffer2))
        {

	    gst_buffer_unref(aacdec_info->inbuffer1);
	    aacdec_info->inbuffer1 = NULL;
	}  
       if (aacdec_info->inbuffer2) {

	    gst_buffer_unref(aacdec_info->inbuffer2);
	    aacdec_info->inbuffer1 = NULL;
	}            
		mfw_gst_aacdec_memclean(aacdec_info);
        
	aacdec_info->total_frames = 0;


	break;
    case GST_STATE_CHANGE_READY_TO_NULL:
	GST_DEBUG("GST_STATE_CHANGE_READY_TO_NULL \n");
	aacdec_global_ptr = NULL;
	break;
    default:
	break;
    }
    GST_DEBUG(" out of mfw_gst_aacdec_change_state routine \n");

    return ret;

}

/*=============================================================================
FUNCTION: mfw_gst_aacdec_get_query_types

DESCRIPTION: gets the different types of query supported by the plugin

ARGUMENTS PASSED:
        pad     - pad on which the function is registered 

RETURN VALUE:
        query types ssupported by the plugin

PRE-CONDITIONS:
        None

POST-CONDITIONS:


IMPORTANT NOTES:
        None
=============================================================================*/
static const GstQueryType *mfw_gst_aacdec_get_query_types(GstPad * pad)
{
    static const GstQueryType src_query_types[] = {
	GST_QUERY_POSITION,
	GST_QUERY_DURATION,
	GST_QUERY_CONVERT,
	0
    };

    return src_query_types;
}

/*==================================================================================================

FUNCTION:   mfw_gst_aacdec_src_query   

DESCRIPTION:    performs query on src pad.    

ARGUMENTS PASSED:
        pad     -   pointer to GstPad
        query   -   pointer to GstQuery        
            
RETURN VALUE:
        TRUE    -   success
        FALSE   -   failure

PRE-CONDITIONS:
        None

POST-CONDITIONS:
		None

IMPORTANT NOTES:
        None

==================================================================================================*/
static gboolean mfw_gst_aacdec_src_query(GstPad * pad,
					     GstQuery * query)
{
    gboolean res = TRUE;
    GstPad *peer;

    MFW_GST_AACDEC_INFO_T *aacdec_info;
    aacdec_info = MFW_GST_AACDEC(GST_OBJECT_PARENT(pad));

    peer = gst_pad_get_peer(aacdec_info->sinkpad);

    switch (GST_QUERY_TYPE(query)) {

    case GST_QUERY_DURATION:
	{

	    GST_DEBUG("coming in GST_QUERY_DURATION \n");
	    GstFormat format;
	    GstFormat rformat;
	    gint64 total, total_bytes;
	    GstPad *peer;

	    /* save requested format */
	    gst_query_parse_duration(query, &format, NULL);
	    if ((peer =
		 gst_pad_get_peer(aacdec_info->sinkpad)) == NULL)
		goto error;

	    if (format == GST_FORMAT_TIME && gst_pad_query(peer, query)) {
		gst_query_parse_duration(query, NULL, &total);
		GST_DEBUG_OBJECT(aacdec_info,
				 "peer returned duration %"
				 GST_TIME_FORMAT, GST_TIME_ARGS(total));

	    }
	    /* query peer for total length in bytes */
	    gst_query_set_duration(query, GST_FORMAT_BYTES, -1);


	    if (!gst_pad_query(peer, query)) {
		goto error;
	    }
	    gst_object_unref(peer);

	    /* get the returned format */
	    gst_query_parse_duration(query, &rformat, &total_bytes);

	    if (rformat == GST_FORMAT_BYTES) {
		GST_DEBUG("peer pad returned total bytes=%d", total_bytes);
	    } else if (rformat == GST_FORMAT_TIME) {
		GST_DEBUG("peer pad returned total time=%",
			  GST_TIME_FORMAT, GST_TIME_ARGS(total_bytes));
	    }

	    /* Check if requested format is returned format */
	    if (format == rformat)
		return TRUE;


	    if (total_bytes != -1) {
		if (format != GST_FORMAT_BYTES) {
		    if (!mfw_gst_aacdec_convert_sink
			(pad, GST_FORMAT_BYTES, total_bytes, &format,
			 &total))
			goto error;
		} else {
		    total = total_bytes;
		}
	    } else {
		total = -1;
	    }
	    aacdec_info->total_time = total;
	    gst_query_set_duration(query, format, total);

	    if (format == GST_FORMAT_TIME) {
		GST_DEBUG("duration=%" GST_TIME_FORMAT,
			  GST_TIME_ARGS(total));
	    } else {
		GST_DEBUG("duration=%" G_GINT64_FORMAT ",format=%u", total,
			  format);
	    }
	    break;
	}
    case GST_QUERY_CONVERT:
	{
	    GstFormat src_fmt, dest_fmt;
	    gint64 src_val, dest_val;
	    gst_query_parse_convert(query, &src_fmt, &src_val, &dest_fmt,
				    &dest_val);
	    if (!
		(res =
		 mfw_gst_aacdec_convert_src(pad, src_fmt, src_val,
						&dest_fmt, &dest_val)))
		goto error;

	    gst_query_set_convert(query, src_fmt, src_val, dest_fmt,
				  dest_val);
	    break;
	}
    default:
	res = FALSE;
	break;
    }
    return res;

  error:
    GST_ERROR("error handling query");
    return FALSE;
}

/*==================================================================================================
FUNCTION:   mfw_gst_aacdec_convert_src   

DESCRIPTION:    converts the format of value from src format to destination format on src pad .    

ARGUMENTS PASSED:
        pad         -   pointer to GstPad   
        src_format  -   format of source value
        src_value   -   value of soure 
        dest_format -   format of destination value
        dest_value  -   value of destination         

RETURN VALUE:
        TRUE    -   sucess
        FALSE   -   failure  

PRE-CONDITIONS:
        None

POST-CONDITIONS:
		None

IMPORTANT NOTES:
        None

==================================================================================================*/
static gboolean
mfw_gst_aacdec_convert_src(GstPad * pad, GstFormat src_format,
			       gint64 src_value, GstFormat * dest_format,
			       gint64 * dest_value)
{
    gboolean res = TRUE;
    guint scale = 1;
    gint bytes_per_sample;

    MFW_GST_AACDEC_INFO_T *aacdec_info;
    aacdec_info = MFW_GST_AACDEC(GST_OBJECT_PARENT(pad));

    bytes_per_sample = aacdec_info->number_of_channels * 4;

    switch (src_format) {
    case GST_FORMAT_BYTES:
	switch (*dest_format) {
	case GST_FORMAT_DEFAULT:
	    if (bytes_per_sample == 0)
		return FALSE;
	    *dest_value = src_value / bytes_per_sample;
	    break;
	case GST_FORMAT_TIME:
	    {
		gint byterate = bytes_per_sample *
		    aacdec_info->sampling_freq;
		if (byterate == 0)
		    return FALSE;
		*dest_value = src_value * GST_SECOND / byterate;
		break;
	    }
	default:
	    res = FALSE;
	}
	break;
    case GST_FORMAT_DEFAULT:
	switch (*dest_format) {
	case GST_FORMAT_BYTES:
	    *dest_value = src_value * bytes_per_sample;
	    break;
	case GST_FORMAT_TIME:
	    if (aacdec_info->sampling_freq == 0)
		return FALSE;
	    *dest_value = src_value * GST_SECOND /
		aacdec_info->sampling_freq;
	    break;
	default:
	    res = FALSE;
	}
	break;
    case GST_FORMAT_TIME:
	switch (*dest_format) {
	case GST_FORMAT_BYTES:
	    scale = bytes_per_sample;
	    /* fallthrough */
	case GST_FORMAT_DEFAULT:
	    *dest_value = src_value * scale *
		aacdec_info->sampling_freq / GST_SECOND;
	    break;
	default:
	    res = FALSE;
	}
	break;
    default:
	res = FALSE;
    }

    return res;
}

/*==================================================================================================

FUNCTION:   mfw_gst_aacdec_convert_sink    

DESCRIPTION:    converts the format of value from src format to destination format on sink pad .  
   

ARGUMENTS PASSED:
        pad         -   pointer to GstPad   
        src_format  -   format of source value
        src_value   -   value of soure 
        dest_format -   format of destination value
        dest_value  -   value of destination 

RETURN VALUE:
        TRUE    -   sucess
        FALSE   -   failure

PRE-CONDITIONS:
        None

POST-CONDITIONS:
		None

IMPORTANT NOTES:
        None

==================================================================================================*/
static gboolean
mfw_gst_aacdec_convert_sink(GstPad * pad, GstFormat src_format,
				gint64 src_value, GstFormat * dest_format,
				gint64 * dest_value)
{
    gboolean res = TRUE;
    float avg_bitrate = 0;
    MFW_GST_AACDEC_INFO_T *aacdec_info;
    GST_DEBUG(" in mfw_gst_aacdec_convert_sink \n");
    aacdec_info = MFW_GST_AACDEC(GST_OBJECT_PARENT(pad));

    if (aacdec_info->app_params.App_adif_header_present){
        avg_bitrate = aacdec_info->app_params.BlockParams.BitRate;
    }
    else  
        avg_bitrate = aacdec_info->bit_rate;

    switch (src_format) {
    case GST_FORMAT_BYTES:
	switch (*dest_format) {
	case GST_FORMAT_TIME:
	    if (avg_bitrate) {

		*dest_value =
		    gst_util_uint64_scale(src_value, 8 * GST_SECOND,
					  avg_bitrate);
	    } else {
		*dest_value = GST_CLOCK_TIME_NONE;
	    }
	    break;
	default:
	    res = FALSE;
	}
	break;
    case GST_FORMAT_TIME:
	switch (*dest_format) {
	case GST_FORMAT_BYTES:
	    if (avg_bitrate) {

		*dest_value = gst_util_uint64_scale(src_value, avg_bitrate,
						    8 * GST_SECOND);

	    } else {
		*dest_value = 0;
	    }

	    break;

	default:
	    res = FALSE;
	}
	break;
    default:
	res = FALSE;
    }

    GST_DEBUG(" out of mfw_gst_aacdec_convert_sink \n");
    return res;
}

/*==================================================================================================

FUNCTION:   mfw_gst_aacdec_seek  

DESCRIPTION:    performs seek operation    

ARGUMENTS PASSED:
        aacdec_info -   pointer to decoder element
        pad         -   pointer to GstPad
        event       -   pointer to GstEvent

RETURN VALUE:
        TRUE    -   sucess
        FALSE   -   failure        

PRE-CONDITIONS:
        None

POST-CONDITIONS:
		None

IMPORTANT NOTES:
        None

==================================================================================================*/
static gboolean
mfw_gst_aacdec_seek(MFW_GST_AACDEC_INFO_T * aacdec_info,
			GstPad * pad, GstEvent * event)
{
    gdouble rate;
    GstFormat format, conv;
    GstSeekFlags flags;
    GstSeekType cur_type, stop_type;
    gint64 cur = 0, stop = 0;
    gint64 time_cur = 0, time_stop = 0;
    gint64 bytes_cur = 0, bytes_stop = 0;
    gboolean flush;
    gboolean res;
    guint bytesavailable;
    gst_event_parse_seek(event, &rate, &format, &flags, &cur_type, &cur,
			 &stop_type, &stop);

    GST_DEBUG("\nseek from  %" GST_TIME_FORMAT "--------------- to %"
	      GST_TIME_FORMAT, GST_TIME_ARGS(cur), GST_TIME_ARGS(stop));

    if (format != GST_FORMAT_TIME) {
	conv = GST_FORMAT_TIME;
	if (!mfw_gst_aacdec_convert_src
	    (pad, format, cur, &conv, &time_cur))
	    goto convert_error;
	if (!mfw_gst_aacdec_convert_src
	    (pad, format, stop, &conv, &time_stop))
	    goto convert_error;
    } else {
	time_cur = cur;
	time_stop = stop;
    }
    GST_DEBUG("\nseek from  %" GST_TIME_FORMAT "--------------- to %"
	      GST_TIME_FORMAT, GST_TIME_ARGS(time_cur),
	      GST_TIME_ARGS(time_stop));

    /* shave off the flush flag, we'll need it later */
    flush = ((flags & GST_SEEK_FLAG_FLUSH) != 0);
    res = FALSE;
    conv = GST_FORMAT_BYTES;



    if (!mfw_gst_aacdec_convert_sink
	(pad, GST_FORMAT_TIME, time_cur, &conv, &bytes_cur))
	goto convert_error;

    if (!mfw_gst_aacdec_convert_sink
	(pad, GST_FORMAT_TIME, time_stop, &conv, &bytes_stop))
	goto convert_error;

    {
	GstEvent *seek_event;


	seek_event =
	    gst_event_new_seek(rate, GST_FORMAT_BYTES, flags, cur_type,
			       bytes_cur, stop_type, bytes_stop);

	/* do the seek */
	res = gst_pad_push_event(aacdec_info->sinkpad, seek_event);

    }


    return TRUE;

    /* ERRORS */
  convert_error:
    {
	/* probably unsupported seek format */
	GST_ERROR("failed to convert format %u into GST_FORMAT_TIME",
		  format);
	return FALSE;
    }
}


/*=============================================================================
FUNCTION:   mfw_gst_aacdec_src_event

DESCRIPTION: This functions handles the events that triggers the
			 source pad of the mpeg4 decoder element.

ARGUMENTS PASSED:
        pad        -    pointer to pad
        event      -    pointer to event
RETURN VALUE:
        TRUE       -	if event is sent to src properly
	    FALSE	   -	if event is not sent to src properly

PRE-CONDITIONS:
        None

POST-CONDITIONS:
        None

IMPORTANT NOTES:
        None
=============================================================================*/
static gboolean
mfw_gst_aacdec_src_event(GstPad * pad, GstEvent * event)
{

    gboolean res;
    MFW_GST_AACDEC_INFO_T *aacdec_info;
    GST_DEBUG(" in mfw_gst_aacdec_src_event routine \n");
    aacdec_info = MFW_GST_AACDEC(GST_OBJECT_PARENT(pad));

    switch (GST_EVENT_TYPE(event)) {
    case GST_EVENT_SEEK:
	gst_event_ref(event);
	if (aacdec_info->seek_flag == TRUE) {
	    res = mfw_gst_aacdec_seek(aacdec_info, pad, event);

	} else {
	    res = gst_pad_push_event(aacdec_info->sinkpad, event);

	}
	break;

    default:
	res = gst_pad_push_event(pad, event);
	break;


    }
    GST_DEBUG(" out of mfw_gst_aacdec_src_event routine \n");

    gst_event_unref(event);
    return res;

}

/*=============================================================================
FUNCTION:   mfw_gst_aacdec_sink_event 

DESCRIPTION:

ARGUMENTS PASSED:
        pad        -    pointer to pad
        event      -    pointer to event
RETURN VALUE:
        TRUE       -	if event is sent to sink properly
	    FALSE	   -	if event is not sent to sink properly

PRE-CONDITIONS:
        None

POST-CONDITIONS:
        None

IMPORTANT NOTES:
        None
=============================================================================*/
static gboolean
mfw_gst_aacdec_sink_event(GstPad * pad, GstEvent * event)
{
    gboolean result = TRUE;
    MFW_GST_AACDEC_INFO_T *aacdec_info;
    AACD_RET_TYPE rc;
    GstCaps *src_caps = NULL, *caps = NULL;
    GstBuffer *outbuffer = NULL;
    GstFlowReturn res = GST_FLOW_OK;
    guint8 *inbuffer;
    gint inbuffsize;
    guint64 time_duration = 0;



    aacdec_info = MFW_GST_AACDEC(GST_OBJECT_PARENT(pad));

    GST_DEBUG(" in mfw_gst_aacdec_sink_event function \n");
    switch (GST_EVENT_TYPE(event)) {
    case GST_EVENT_NEWSEGMENT:
	{
	    GstFormat format;
	    gint64 start, stop, position;
	    gint64 nstart, nstop;
	    GstEvent *nevent;

	    GST_DEBUG(" in GST_EVENT_NEWSEGMENT \n");
	    gst_event_parse_new_segment(event, NULL, NULL, &format, &start,
					&stop, &position);

	    if (format == GST_FORMAT_BYTES) {
		format = GST_FORMAT_TIME;
		if (start != 0)
		    result =
			mfw_gst_aacdec_convert_sink(pad,
							GST_FORMAT_BYTES,
							start, &format,
							&nstart);
		else
		    nstart = start;
		if (stop != 0)
		    result =
			mfw_gst_aacdec_convert_sink(pad,
							GST_FORMAT_BYTES,
							stop, &format,
							&nstop);
		else
		    nstop = stop;

		nevent =
		    gst_event_new_new_segment(FALSE, 1.0, GST_FORMAT_TIME,
					      nstart, nstop, nstart);
		gst_event_unref(event);
		aacdec_info->time_offset = (guint64) nstart;

		result =
		    gst_pad_push_event(aacdec_info->srcpad, nevent);
		if (TRUE != result) {
		    GST_ERROR
			("\n Error in pushing the event,result	is %d\n",
			 result);

		}
	    } else if (format == GST_FORMAT_TIME) {
		aacdec_info->time_offset = (guint64) start;

		result =
		    gst_pad_push_event(aacdec_info->srcpad, event);
		if (TRUE != result) {
		    GST_ERROR
			("\n Error in pushing the event,result	is %d\n",
			 result);
		   
		}
	    }
	    break;
	}
    case GST_EVENT_EOS:
	{

	    GST_DEBUG("\nDecoder: Got an EOS from Demuxer\n");
	    if (aacdec_global_ptr == NULL)
		aacdec_global_ptr = aacdec_info;
	    else {
		if (aacdec_global_ptr != aacdec_info) {
		    gboolean res;
		    res =
			gst_pad_push_event(aacdec_info->srcpad, event);
		    if (res != TRUE) {
			GST_ERROR("aac dec:could not push onto \
                            next element %d\n", res);
		    }
		    return TRUE;
		}
	    }

	    if (aacdec_info->init_done) {

		inbuffsize = GST_BUFFER_SIZE(aacdec_info->inbuffer1);
		aacdec_info->eos = TRUE;

        
            aacdec_info->inbuffer2=NULL;
		result =
		    mfw_gst_aacdec_data(aacdec_info, &inbuffsize);
        

		if (result != TRUE) {
		    GST_ERROR("Error in decoding the frame");
		}


	    }

	    result = gst_pad_push_event(aacdec_info->srcpad, event);
	    if (TRUE != result) {
		GST_ERROR
		    ("\n Error in pushing the event,result	is %d\n",
		     result);
	
	    }

            return TRUE;
	    break;
	}
    case GST_EVENT_FLUSH_STOP:
	{

	    GST_DEBUG(" GST_EVENT_FLUSH_STOP \n");

	    result = mfw_gst_aacdec_mem_flush(aacdec_info);

	    if (aacdec_info->inbuffer2) {
		gst_buffer_unref(aacdec_info->inbuffer2);
		aacdec_info->inbuffer2 = NULL;
	    }
	    result = gst_pad_push_event(aacdec_info->srcpad, event);
	    if (TRUE != result) {
		GST_ERROR("\n Error in pushing the event,result	is %d\n",
			  result);
	
	    }
	    break;
	}

    case GST_EVENT_FLUSH_START:
    default:
	{
	    result = gst_pad_event_default(pad, event);
	    break;
	}
    }

    GST_DEBUG(" out of mfw_gst_aacdec_sink_event \n");

    return result;
}


/*=============================================================================
FUNCTION:   mfw_gst_aacdec_set_caps

DESCRIPTION:    this function handles the link with other plug-ins and used for
                capability negotiation  between pads

ARGUMENTS PASSED:
        pad        -    pointer to GstPad
        caps       -    pointer to GstCaps

RETURN VALUE:
        TRUE       -    if capabilities are set properly
        FALSE      -    if capabilities are not set properly
PRE-CONDITIONS:
        None

POST-CONDITIONS:
        None

IMPORTANT NOTES:
        None
=============================================================================*/
static gboolean mfw_gst_aacdec_set_caps(GstPad * pad, GstCaps * caps)
{



    MFW_GST_AACDEC_INFO_T *aacdec_info;
    const gchar *mime;
    gint mpeg_version;
    GstStructure *structure = gst_caps_get_structure(caps, 0);
    aacdec_info = MFW_GST_AACDEC(gst_pad_get_parent(pad));

    GST_DEBUG(" in mfw_gst_aacdec_set_caps routine \n");
    mime = gst_structure_get_name(structure);

    if (strcmp(mime, "audio/mpeg") != 0) {
	GST_WARNING
	    ("Wrong	mimetype %s	provided, we only support %s",
	     mime, "audio/mpeg");
	return FALSE;
    }
    gst_structure_get_int(structure, "mpegversion", &mpeg_version);
    if (((mpeg_version != 2) && (mpeg_version != 4))) {
	GST_ERROR("Caps negotiation: mpeg version not supprted");
	return FALSE;
    }

    gst_structure_get_int(structure, "bitrate",
			  &aacdec_info->bit_rate);

    if (!gst_pad_set_caps(aacdec_info->srcpad, caps)) {
	return FALSE;
    }

    GST_DEBUG(" out of mfw_gst_aacdec_set_caps routine \n");

    return TRUE;
}

/*=============================================================================
FUNCTION:   mfw_gst_aacdec_init

DESCRIPTION:This function creates the pads on the elements and register the
			function pointers which operate on these pads.

ARGUMENTS PASSED:
        pointer the aac decoder element handle.

RETURN VALUE:
        None

PRE-CONDITIONS:
        _base_init and _class_init are called

POST-CONDITIONS:
        None

IMPORTANT NOTES:
        None
=============================================================================*/

static void
mfw_gst_aacdec_init(MFW_GST_AACDEC_INFO_T * aacdec_info)
{

    GstElementClass *klass = GST_ELEMENT_GET_CLASS(aacdec_info);
    GST_DEBUG(" \n in mfw_gst_aacdec_init routine \n");
    aacdec_info->sinkpad =
	gst_pad_new_from_template(gst_element_class_get_pad_template
				  (klass, "sink"), "sink");

    aacdec_info->srcpad =
	gst_pad_new_from_template(gst_element_class_get_pad_template
				  (klass, "src"), "src");

    gst_element_add_pad(GST_ELEMENT(aacdec_info),
			aacdec_info->sinkpad);
    gst_element_add_pad(GST_ELEMENT(aacdec_info),
			aacdec_info->srcpad);

    gst_pad_set_setcaps_function(aacdec_info->sinkpad,
				 mfw_gst_aacdec_set_caps);
    gst_pad_set_chain_function(aacdec_info->sinkpad,
			       mfw_gst_aacdec_chain);

    gst_pad_set_event_function(aacdec_info->sinkpad,
			       GST_DEBUG_FUNCPTR
			       (mfw_gst_aacdec_sink_event));

    gst_pad_set_query_type_function(aacdec_info->srcpad,
				    GST_DEBUG_FUNCPTR
				    (mfw_gst_aacdec_get_query_types));
    gst_pad_set_event_function(aacdec_info->srcpad,
			       GST_DEBUG_FUNCPTR
			       (mfw_gst_aacdec_src_event));


    GST_DEBUG("\n out of mfw_gst_aacdec_init \n");


}

/*=============================================================================
FUNCTION:   mfw_gst_aacdec_class_init

DESCRIPTION:Initialise the class only once (specifying what signals,
            arguments and virtual functions the class has and setting up
            global state)
ARGUMENTS PASSED:
       	klass   - pointer to aac decoder's element class

RETURN VALUE:
        None

PRE-CONDITIONS:
        None

POST-CONDITIONS:
        None

IMPORTANT NOTES:
        None
=============================================================================*/
static void
mfw_gst_aacdec_class_init(MFW_GST_AACDEC_CLASS_T * klass)
{
    GObjectClass *gobject_class = NULL;
    GstElementClass *gstelement_class = NULL;
    gobject_class = (GObjectClass *) klass;
    gstelement_class = (GstElementClass *) klass;
    parent_class_aac =
	(GstElementClass *) g_type_class_ref(GST_TYPE_ELEMENT);
    gobject_class->set_property = mfw_gst_aacdec_set_property;
    gobject_class->get_property = mfw_gst_aacdec_get_property;
    gstelement_class->change_state = mfw_gst_aacdec_change_state;
}

/*=============================================================================
FUNCTION:  mfw_gst_aacdec_base_init

DESCRIPTION:
            aac decoder element details are registered with the plugin during
            _base_init ,This function will initialise the class and child
            class properties during each new child class creation


ARGUMENTS PASSED:
        Klass   -   pointer to aac decoder plug-in class

RETURN VALUE:
        None

PRE-CONDITIONS:
        None

POST-CONDITIONS:
        None

IMPORTANT NOTES:
        None
=============================================================================*/

static void
mfw_gst_aacdec_base_init(MFW_GST_AACDEC_CLASS_T * klass)
{
    static GstElementDetails element_details = {
	"Freescale AAC Decoder Plugin",
	"Codec/Decoder/Audio",
	"decodes AAC + LC bitstreams",
	"Multimedia Team <teammmsw@freescale.com>"
    };
    GstElementClass *element_class = GST_ELEMENT_CLASS(klass);
    gst_element_class_add_pad_template(element_class,
				       gst_static_pad_template_get
				       (&src_factory));
    gst_element_class_add_pad_template(element_class,
				       gst_static_pad_template_get
				       (&sink_factory));
    gst_element_class_set_details(element_class, &element_details);
}

/*=============================================================================
FUNCTION: mfw_gst_aacdec_get_type

DESCRIPTION:    intefaces are initiated in this function.you can register one
                or more interfaces  after having registered the type itself.

ARGUMENTS PASSED:
            None

RETURN VALUE:
                 A numerical value ,which represents the unique identifier of this
            element(aacdecoder)

PRE-CONDITIONS:
            None

POST-CONDITIONS:
            None

IMPORTANT NOTES:
            None
=============================================================================*/
GType mfw_gst_aacdec_get_type(void)
{
    static GType aacdec_type = 0;

    if (!aacdec_type) {
	static const GTypeInfo aacdec_info = {
	    sizeof(MFW_GST_AACDEC_CLASS_T),
	    (GBaseInitFunc) mfw_gst_aacdec_base_init,
	    NULL,
	    (GClassInitFunc) mfw_gst_aacdec_class_init,
	    NULL,
	    NULL,
	    sizeof(MFW_GST_AACDEC_INFO_T),
	    0,
	    (GInstanceInitFunc) mfw_gst_aacdec_init,
	};
	aacdec_type = g_type_register_static(GST_TYPE_ELEMENT,
						 "MFW_GST_AACDEC_INFO_T",
						 &aacdec_info,
						 (GTypeFlags) 0);
    }
    GST_DEBUG_CATEGORY_INIT(mfw_gst_aacdec_debug, "mfw_aacdecoder",
			    0, "FreeScale's AAC Decoder's Log");

    return aacdec_type;
}


/*****************************************************************************/
/*    This is used to define the entry point and meta data of plugin         */
/*****************************************************************************/
GST_PLUGIN_DEFINE(GST_VERSION_MAJOR,	/* major version of Gstreamer    */
		  GST_VERSION_MINOR,	/* minor version of Gstreamer    */
		  "mfw_aacdecoder",	/* name of the plugin            */
		  "decodes AAC + LC streams ",	/* what plugin actually does     */
		  plugin_init,	/* first function to be called   */
		  VERSION,
		  GST_LICENSE_UNKNOWN,
		  "freescale semiconductor", "www.freescale.com ")
/*=============================================================================
FUNCTION:   plugin_init

DESCRIPTION:    special function , which is called as soon as the plugin or
                element is loaded and information returned by this function
                will be cached in central registry

ARGUMENTS PASSED:
        plugin     -    pointer to container that contains features loaded
                        from shared object module

RETURN VALUE:
        return TRUE or FALSE depending on whether it loaded initialized any
        dependency correctly

PRE-CONDITIONS:
        None

POST-CONDITIONS:
        None

IMPORTANT NOTES:
        None
=============================================================================*/
static gboolean plugin_init(GstPlugin * plugin)
{
    return gst_element_register(plugin, "mfw_aacdecoder",
				GST_RANK_PRIMARY, MFW_GST_TYPE_AACDEC);
}

/*=============================================================================
FUNCTION: mfw_gst_aacdec_mem_flush

DESCRIPTION: this function flushes the current memory and allocate the new memory
                for decoder . 

ARGUMENTS PASSED:
        mp3dec_info -   pointer to mp3decoder element structure      

RETURN VALUE:
        None

PRE-CONDITIONS:
        None

POST-CONDITIONS:


IMPORTANT NOTES:
        None
=============================================================================*/

static gboolean mfw_gst_aacdec_mem_flush(MFW_GST_AACDEC_INFO_T *
					     aacdec_info)
{
    gint loopctr = 0;
    gboolean result = TRUE;
    gint num;
    gint rec_no = 0;
    gint nr;
    AACD_Decoder_Config *dec_config = NULL;
    AACD_Mem_Alloc_Info_Sub *mem;
    AACD_RET_TYPE rc = 0;

    GST_DEBUG("in mfw_gst_aacdec_mem_flush \n");
    if (!aacdec_info->init_done)
	return FALSE;


    
    mfw_gst_aacdec_memclean(aacdec_info);

    memset(&aacdec_info->app_params, 0, sizeof(AACD_App_params *));

   /* allocate memory for config structure */
	dec_config = (AACD_Decoder_Config *)
	    aacd_alloc_fast(sizeof(AACD_Decoder_Config));
	aacdec_info->app_params.dec_config = dec_config;
	if (dec_config == NULL) {
	    GST_ERROR("error in allocation of decoder config structure");
	    return FALSE;
	}



	/* call query mem function to know mem requirement of library */

	if (aacd_query_dec_mem(dec_config) != AACD_ERROR_NO_ERROR) {
	    GST_ERROR
		("Failed to get the memory configuration for the decoder\n");
	    return FALSE;
	}

	/* Number of memory chunk requests by the decoder */
	nr = dec_config->aacd_mem_info.aacd_num_reqs;

	 for(rec_no = 0; rec_no < nr; rec_no++)
     {
         mem = &(dec_config->aacd_mem_info.mem_info_sub[rec_no]);
         
         if (mem->aacd_type == AACD_FAST_MEMORY)
         {
             mem->app_base_ptr = aacd_alloc_fast (mem->aacd_size);
             if (mem->app_base_ptr == NULL)
                 return FALSE;
         }
         else
         {
             mem->app_base_ptr = aacd_alloc_slow (mem->aacd_size);
             if (mem->app_base_ptr == NULL)
                 return FALSE;
         }
         memset(dec_config->aacd_mem_info.mem_info_sub[rec_no].app_base_ptr,
             0, dec_config->aacd_mem_info.mem_info_sub[rec_no].aacd_size); 
         
     }

	aacdec_info->app_params.BitsInHeader = 0;
	aacdec_info->app_params.App_adif_header_present = FALSE;

	/* register the call-back function in the decoder context */
	dec_config->app_swap_buf = app_swap_buffers_aac_dec;

#ifndef OUTPUT_24BITS
            dec_config->num_pcm_bits     = AACD_16_BIT_OUTPUT;
#else
            dec_config->num_pcm_bits     = AACD_24_BIT_OUTPUT;
#endif  /*OUTPUT_24BITS*/				/*OUTPUT_24BITS */


	rc = aacd_decoder_init(dec_config);
	if (rc != AACD_ERROR_NO_ERROR) {
	    GST_ERROR("Error in initializing the decoder");
	    return FALSE;
	}

    aacdec_info->init_done = FALSE;

    GST_DEBUG("out of mfw_gst_aacdec_mem_flush\n");
    return result;
}
